import inspect
from ..task import Task
from .pickle_task import unpickle_task


class PickledTask(Task):
    async def run(self, func: str, **inputs):
        func = unpickle_task(func)
        if inspect.iscoroutinefunction(func):
            return await func(**inputs)
        else:
            return func(**inputs)



import docker

# shim that allows passing dockerfiles as a string together with a context path.
# this removes the need to write temporary files!
# source: https://github.com/docker/docker-py/issues/2105
docker.api.build.process_dockerfile = lambda dockerfile, path: ('Dockerfile', dockerfile)


class Dockerfile(object):
    """ Simple tool for creating dockerfiles """

    def __init__(self, base):
        self.lines = [f'FROM {base}']

    def copy(self, src, dst):
        self.lines.append(f'COPY {src} {dst}')

    def run(self, cmd):
        self.lines.append(f'RUN {cmd}')

    def env(self, key: str, value: str):
        self.lines.append(f'ENV {key}={value}')

    def workdir(self, path):
        self.lines.append(f'WORKDIR {path}')

    def __str__(self):
        return '\n'.join(self.lines)

    @staticmethod
    def read(path):
        with open(path, 'r') as f:
            df = Dockerfile('none')
            df.lines = f.readlines()
            return df



from random import randint
from cowait.worker.worker_node import WorkerNode
from .html_logger import HTMLLogger


class NotebookNode(WorkerNode):
    """
    The Notebook Node is a variant of the standard worker node meant to run in a notebook.
    It simulates a running task by connecting upstream and forwarding events from tasks
    created from within the notebook.

    NotebookNodes use random ports for their web servers to allow multiple nodes on a single host.
    Output is disabled to prevent event spam into the notebook.
    """

    def __init__(self, taskdef):
        self.taskdef = taskdef
        super().__init__(
            id=taskdef.id,
            upstream=taskdef.upstream,
            port=randint(10000, 60000),
            logger=HTMLLogger(),
        )

    async def start(self, token: str) -> None:
        """
        Starts the node by connecting upstream, sending initialization
        events and starting the local web server.
        """
        await self.connect(token)
        await self.parent.send_init(self.taskdef)
        await self.parent.send_run()
        await self.parent.send_log(data='Kernel ready.', file='stdout')
        self.serve()

    async def stop(self):
        await self.parent.send_log(data='Kernel stopped!', file='stdout')
        await self.parent.send_stop()
        await self.parent.close()

    async def connect(self, token: str) -> None:
        await self.parent.connect(self.upstream, token)



from ..config import Config
from ..context import Context
from cowait.engine import ProviderError
from .run import RunLogger


def destroy(config: Config, *, cluster_name: str = None) -> None:
    try:
        context = Context.open(config)
        cluster = context.get_cluster(cluster_name)
        cluster.destroy_all()

    except ProviderError as e:
        print('Provider error:', str(e))


def list_tasks(config: Config, *, cluster_name: str = None) -> None:
    try:
        context = Context.open(config)
        cluster = context.get_cluster(cluster_name)
        tasks = cluster.list_all()
        for task in tasks:
            print(task.id)

    except ProviderError as e:
        print('Provider error:', str(e))


def kill(config: Config, partial_task_id: str, cluster_name: str):
    """ Kills all tasks that somewhat matches partial_task_id """
    try:
        context = Context.open(config)
        cluster = context.get_cluster(cluster_name)
        tasks = cluster.list_all()
        for task in tasks:
            if partial_task_id in task.id:
                cluster.destroy(task.id)
                print('killed ', task.id)

    except ProviderError as e:
        print('Provider error:', str(e))


def logs(config: Config, task_id: str, cluster_name: str, raw: bool = False):
    try:
        context = Context.open(config)
        cluster = context.get_cluster(cluster_name)
    
        logs = cluster.logs(task_id)
        logger = RunLogger(raw)
        logger.header('task output')
        for msg in logs:
            logger.handle(msg)
        logger.header()

    except ProviderError as e:
        print('Provider error:', str(e))




from __future__ import annotations
import re
from cowait.version import version

VERSION_FORMAT = re.compile('([0-9]+)\\.([0-9]+)\\.([0-9]+)')


class Version(object):
    def __init__(self, major: int, minor: int, revision: int):
        self.major = major
        self.minor = minor
        self.revision = revision

    def __str__(self) -> str:
        return f'{self.major}.{self.minor}.{self.revision}'

    def is_compatible(self):
        # 0.3 and below are not compatible.
        if self.minor < 4:
            return False
        return True

    @staticmethod
    def current():
        return Version.parse(version)

    @staticmethod
    def parse(version: str) -> Version:
        if re.match(VERSION_FORMAT, version) is None:
            raise ValueError(f'Illegal version string {version}')
        major, minor, rev = map(lambda v: int(v), version.split('.'))
        return Version(major, minor, rev)



from .type import Type
from .mapping import TypeAlias, convert_type
from datetime import datetime


@TypeAlias(any)
class Any(Type):
    """ Any type. Disables typechecking. """

    name: str = 'any'

    def validate(self, value: any, name: str) -> None:
        pass

    def serialize(self, value: any) -> any:
        guessed_type = convert_type(type(value))
        return guessed_type.serialize(value)


@TypeAlias(int)
class Int(Type):
    """ Integer, or anything that can be cast to an integer """

    name: str = 'int'

    def validate(self, value: int, name: str) -> None:
        try:
            int(value)
        except ValueError:
            raise ValueError(f'{name} must be an integer')

    def serialize(self, value: int) -> int:
        return int(value)

    def deserialize(self, value: any) -> int:
        return int(value)


@TypeAlias(float)
class Float(Type):
    """ Floating point, or anything that can be cast to a float """

    name: str = 'float'

    def validate(self, value: float, name: str) -> None:
        try:
            float(value)
        except ValueError:
            raise ValueError(f'{name} must be a float')

    def serialize(self, value: float) -> float:
        return float(value)

    def deserialize(self, value: any) -> float:
        return float(value)


@TypeAlias(str)
class String(Type):
    """ String """

    name: str = 'str'

    def validate(self, value: str, name: str) -> None:
        try:
            str(value)
        except ValueError:
            raise ValueError(f'{name} must be a string')

    def serialize(self, value: str) -> str:
        return str(value)

    def deserialize(self, value: any) -> str:
        return str(value)


@TypeAlias(bool)
class Bool(Type):
    """ Boolean """

    name: str = 'bool'

    def validate(self, value: bool, name: str) -> None:
        if isinstance(value, bool):
            return True

        value = str(value).lower()
        if value == 'true':
            return True
        elif value == 'false':
            return False
        else:
            raise ValueError(f'{name} must be a boolean')

    def serialize(self, value: bool) -> bool:
        return bool(value)

    def deserialize(self, value: any) -> bool:
        if isinstance(value, bool):
            return value

        value = str(value).lower()
        if value == 'true':
            return True
        elif value == 'false':
            return False


@TypeAlias(datetime)
class DateTime(Type):
    """ Python datetime object serialized as an ISO8601 string """

    def validate(self, value: str, name: str) -> None:
        if isinstance(value, datetime):
            return

        if not isinstance(value, str):
            raise ValueError('Expected ISO8601 datetime')

        datetime.fromisoformat(value)

    def serialize(self, value: datetime) -> str:
        return value.isoformat()

    def deserialize(self, value: str) -> datetime:
        return datetime.fromisoformat(value)


@TypeAlias(type(None), None)
class Void(Type):
    def validate(self, value, name: str):
        if value is not None:
            raise ValueError(f'Expected {name} to be None')

    def serialize(self, value):
        return None

    def deserialize(self, value):
        return None



from .type import Type
from .simple import Any
from .mapping import is_cowait_type, convert_type, TypeAlias


@TypeAlias(dict)
class Dict(Type):
    """ Dictionary type """

    def __init__(self, shape: dict = {}):
        for key, type in shape.items():
            if not is_cowait_type(type):
                raise ValueError(f'Key {key} is not a cowait Type')

        self.shape = {
            key: convert_type(type)
            for key, type in shape.items()
        }

    def validate(self, value: dict, name: str) -> None:
        if not isinstance(value, dict):
            raise ValueError(f'{name} is not a dict')

        for key, type in self.shape.items():
            if key not in value:
                raise ValueError(f'{name}[{key}] is required')
            type.validate(value[key], f'{name}[{key}]')

    def serialize(self, value: dict) -> dict:
        any = Any()
        return {
            key: self.shape.get(key, any).serialize(value)
            for key, value in value.items()
        }

    def deserialize(self, value: dict) -> dict:
        any = Any()
        return {
            key: self.shape.get(key, any).deserialize(value)
            for key, value in value.items()
        }

    def describe(self):
        desc = {}
        for key, type in self.shape.items():
            type_desc = type.describe()
            if type_desc is not None:
                desc[key] = type_desc
        return desc

    def __getitem__(self, key):
        return self.shape[key]



import re
from cowait.tasks.shell import ShellTask
from cowait.engine import env_pack, \
    ENV_TASK_CLUSTER, ENV_TASK_DEFINITION, ENV_GZIP_ENABLED
from .kernel import ENV_KERNEL_TOKEN

TOKEN_PATTERN = re.compile('\\?token\\=([a-z0-9]+)')


class NotebookTask(ShellTask):
    def init(self):
        self.jupyter_token = None

    async def run(self, **inputs):
        print('Starting JupyterLab...')

        await super().run(
            command='jupyter lab --no-browser',
            env={
                ENV_GZIP_ENABLED: '1',
                ENV_TASK_DEFINITION: env_pack(self.taskdef.serialize()),
                ENV_TASK_CLUSTER: env_pack(self.cluster.serialize()),
                ENV_KERNEL_TOKEN: self.node.server.auth.get_token(),
            },
        )

    def on_ready(self):
        print()
        print('JupyterLab ready!')
        if '/' in self.taskdef.routes:
            url = self.taskdef.routes['/']['url']
            print('Notebook available at:')
            print(f'  {url}?token={self.jupyter_token}')
        else:

            print('Warning: No route set')

    def filter_stdout(self, line):
        return self.filter_jupyter_token(line)

    def filter_stderr(self, line):
        return self.filter_jupyter_token(line)

    def filter_jupyter_token(self, line):
        if self.jupyter_token is None:
            match = re.search(TOKEN_PATTERN, line)
            if match is not None:
                self.jupyter_token = match[1]
                self.on_ready()
        return False



import click
import cowait.cli.commands
from .utils import parse_input_list


@click.command(help='build a task')
@click.option('-q', '--quiet',
              type=bool, is_flag=True,
              help='no output except result',
              default=False)
@click.option('-w', '--workdir',
              default=None,
              type=str,
              help='task working directory')
@click.option('-i', '--image',
              default=None,
              type=str,
              help='image name')
@click.option('-a', '--arg',
              type=str,
              multiple=True,
              help='docker build argument')
@click.option('-p', '--push',
              type=bool, is_flag=True,
              help='push image after building',
              default=False)
@click.pass_context
def build(ctx, quiet: bool, workdir: str, image: str, arg: dict, push: bool):
    cowait.cli.build(
        ctx.obj,
        quiet=quiet,
        workdir=workdir,
        image_name=image,
        buildargs=parse_input_list(arg),
        push=push,
    )


@click.command(help='push a task to the registry')
@click.pass_context
def push(ctx):
    cowait.cli.push(ctx.obj)



import sys
from cowait.tasks import TaskDefinition, TASK_LOG
from cowait.engine import ProviderError, TaskCreationError
from ..config import Config
from ..context import Context
from ..utils import ExitTrap
from ..logger import Logger
from .build import build as run_build
from .push import push as run_push


def test(
    config: Config,
    push: bool,
    cluster_name: str = None,
):
    logger = TestLogger()
    try:
        context = Context.open(config)
        cluster = context.get_cluster(cluster_name)

        if push:
            run_push()
        else:
            run_build(config)

        # execute the test task within the current image
        task = cluster.spawn(TaskDefinition(
            name='cowait.test',
            image=context.image,
        ))

        def destroy(*args):
            logger.header('interrupt')
            cluster.destroy(task.id)
            sys.exit(1)

        with ExitTrap(destroy):
            # capture & print logs
            logs = cluster.logs(task.id)
            logger.header('task output')
            for msg in logs:
                logger.handle(msg)

        logger.header()

        # grab task result
        passing = cluster.wait(task)
        sys.exit(0 if passing else 1)

    except ProviderError as e:
        logger.print_exception(f'Provider Error: {e}')
        sys.exit(1)

    except TaskCreationError as e:
        logger.print_exception(f'Error creating task: {e}')
        sys.exit(1)


class TestLogger(Logger):
    def handle(self, msg):
        type = msg.get('type', None)
        if type == TASK_LOG:
            print(msg['data'], end='')



import sys
from cowait.cli import CliError
from cowait.cli.config import Config
from cowait.cli.app import new_cli_app


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    config = Config.get_global()
    app = new_cli_app()
    try:
        app(obj=config)
    except CliError as e:
        print(f'Error: {e}')


if __name__ == "__main__":
    sys.exit(main())



import click
import cowait.cli


@click.command(help='create a new context')
@click.argument('name', type=str, required=False)
@click.option('--image', type=str, required=False, help='image name')
@click.option('--base', type=str, required=False, help='base image name')
@click.pass_context
def new(ctx, name: str, image: str, base: str):
    cowait.cli.new_context(
        ctx.obj,
        name=name,
        image=image,
        base=base,
    )



import json
from cowait.tasks import TaskDefinition
from cowait.engine.const import ENV_TASK_DEFINITION
from cowait.engine.utils import env_unpack
from .errors import PodTerminatedError, PodUnschedulableError, ImagePullError


def pod_is_ready(pod):
    if pod.status.conditions:
        for cond in pod.status.conditions:
            if cond.type == 'PodScheduled' and cond.status == 'False':
                raise PodUnschedulableError(cond.message)

    statuses = pod.status.container_statuses
    if statuses is None:
        return False

    for status in statuses:
        state = status.state

        if state.running is not None:
            return True

        if state.terminated is not None:
            raise PodTerminatedError()

        if state.waiting is not None:
            if state.waiting.reason == 'ContainerCreating':
                return False

            if state.waiting.reason == 'ErrImagePull':
                raise ImagePullError(state.waiting.message)

            if state.waiting.reason == 'ImagePullBackoff':
                raise ImagePullError(state.waiting.message)

    return False


def extract_pod_taskdef(pod) -> TaskDefinition:
    for container in pod.spec.containers:
        for env in container.env:
            if env.name == ENV_TASK_DEFINITION:
                taskdef = env_unpack(env.value)
                return TaskDefinition(**taskdef)
    raise Exception('Failed to extract pod task definition')




import os.path
import docker
import docker.errors
import docker.credentials.errors
from cowait.utils.const import DEFAULT_BASE_IMAGE
from ..task_image import TaskImage, BuildError
from ..docker_file import Dockerfile
from ..context import Context, CONTEXT_FILE_NAME
from ..config import Config
from ..logger import Logger
from .push import push as push_command


def build(
    config: Config, *,
    image_name: str = None,
    workdir: str = None,
    buildargs: dict = {},
    quiet: bool = False,
    push: bool = False,
) -> TaskImage:
    logger = Logger(quiet)
    try:
        if not Context.exists():
            logger.println(f'No {CONTEXT_FILE_NAME} found. '
                           f'Using default image: {DEFAULT_BASE_IMAGE}')
            return TaskImage.get(DEFAULT_BASE_IMAGE)

        context = Context.open(config)
        context.override('workdir', workdir)

        if image_name is not None:
            context.override('image', image_name)

        image = TaskImage.open(context)
        logger.header('BUILD')
        logger.println('Image:', image.name)
        logger.println('Context Root:', context.root_path)
        logger.println('Workdir:', context.workdir)

        if len(buildargs) > 0:
            logger.println('Buildargs:')
            for arg in buildargs.keys():
                logger.println(f'  {arg}')

        # find task-specific requirements.txt
        # if it exists, it will be copied to the container, and installed
        # custom requirements are ignored when using custom dockerfiles.
        requirements = context.file_rel('requirements.txt')
        if requirements:
            logger.println('* Found custom requirements.txt')

        # find custom Dockerfile
        # if it exists, build and extend that instead of the default base image
        base_image = context.base
        dockerfile = context.file('Dockerfile')
        if dockerfile:
            logger.println('* Found custom Dockerfile:', context.relpath(dockerfile))

            # cant override base image when using custom dockerfiles
            if base_image != DEFAULT_BASE_IMAGE:
                logger.println('! Warning: Base image setting is ignored when using a custom dockerfile')

            # disable automatic requirements.txt install when using custom dockerfiles.
            # notify user to avoid confusion
            if requirements:
                logger.println('! Warning: requirements.txt is not automatically installed when using custom dockerfiles.')
                requirements = None

            if not os.path.isfile('.dockerignore'):
                logger.println('! Warning: No .dockerignore file found.')

            logger.header('BASE')
            basedf = Dockerfile.read(dockerfile)
            base = TaskImage.build_image(
                path=os.path.dirname(dockerfile),
                dockerfile=str(basedf),
                quiet=quiet,
                buildargs=buildargs,
            )
            if base is None:
                raise RuntimeError('Failed to build base image')
            base_image = base.id

        logger.header('IMAGE')
        image.build(
            base=base_image,
            requirements=requirements,
            quiet=quiet,
            buildargs=buildargs,
        )

        logger.header()

        if push:
            push_command(config)

        return image

    except docker.errors.DockerException as e:
        logger.print_exception(f'Docker exception: {e}')

    except BuildError as e:
        logger.print_exception(str(e))



import os.path
from cowait.engine import get_cluster_provider
from .const import CONTEXT_FILE_NAME
from .settings_dict import SettingsDict


def get_global_config_path():
    home = os.path.expanduser('~')
    return f'{home}/.{CONTEXT_FILE_NAME}'


class Config(SettingsDict):
    def __init__(
        self, *,
        path: str = None,
        data: dict = None,
        parent: 'Config' = None,
    ):
        super().__init__(path=path, data=data, parent=parent)

    @property
    def clusters(self) -> list:
        return self.get('clusters', {}, False)

    @property
    def default_cluster(self) -> str:
        return self.get('default_cluster', 'docker', False)

    @default_cluster.setter
    def default_cluster(self, value):
        self.set("default_cluster", value)

    def get_cluster(self, cluster_name: str = None):
        if cluster_name is None:
            cluster_name = self.default_cluster
        if cluster_name not in self.clusters:
            raise RuntimeError(
                f'No configuration found for cluster {cluster_name}')
        return get_cluster_provider(**self.get(f'clusters.{cluster_name}'))

    def write(self) -> None:
        path = get_global_config_path()
        return super().write(path)

    @staticmethod
    def get_local(path: str = None) -> 'Config':
        return Config(path=path, parent=Config.get_global())

    @staticmethod
    def get_global() -> 'Config':
        default = Config.get_default()
        path = get_global_config_path()

        if not os.path.exists(path):
            return Config(data={}, parent=default)

        return Config(path=path, parent=default)

    @staticmethod
    def get_default() -> 'Config':
        return Config(data={
            'default_cluster': 'docker',
            'clusters': {
                'docker': {
                    'type': 'docker',
                    'network': 'cowait',
                },
                'kubernetes': {
                    'type': 'kubernetes',
                },
            },
        })



from cowait.tasks import Task, rpc, sleep


class RpcChild(Task):
    def init(self):
        self.called = False

    async def run(self) -> dict:
        print('waiting for call from parent...')
        for i in range(0, 5):
            if self.called:
                break
            await sleep(1)

        # call parent
        await self.parent.set_called()

        return {
            'called': self.called
        }

    @rpc
    async def echo(self, **params) -> dict:
        print('echo', params)
        self.called = True
        return params



import os
import asyncio
from datetime import datetime
from aiohttp import WSMsgType, ClientError, ClientSession
from cowait.utils import EventEmitter
from .rpc_client import RpcClient
from .const import ON_CONNECT, ON_CLOSE, ON_ERROR
from .errors import AuthError, SocketError


class Client(EventEmitter):
    def __init__(self):
        super().__init__()
        self.ws = None
        self.rpc = None
        self.buffer = []

    @property
    def connected(self) -> bool:
        return self.ws is not None

    async def connect(
        self,
        url: str,
        token: str,
        max_retries: int = 5
    ) -> None:
        retries = 0
        while retries < max_retries or max_retries == 0:
            retries += 1
            try:
                await self._connect(url, token)
            except ClientError as e:
                if hasattr(e, 'status') and e.status == 401:
                    raise AuthError('Upstream authentication failed')
                else:
                    print('Upstream connection failed:', str(e))
                    await asyncio.sleep(retries * 3)
            finally:
                self.ws = None

        if retries >= max_retries and max_retries > 0:
            # Reached maximum number of connection attempts
            # Crash the client
            print('Max upstream connection attempts reached')
            os._exit(1)

    async def _connect(self, url: str, token: str) -> None:
        async with ClientSession() as session:
            async with session.ws_connect(
                url,
                headers={'Authorization': f'Bearer {token}'},
                autoping=True,
                heartbeat=5.0,
                timeout=30.0,
            ) as ws:
                self.ws = ws
                self.rpc = RpcClient(ws)
                await self.emit(ON_CONNECT, conn=self)

                # send buffered messages
                for msg in self.buffer:
                    await self.send(msg)
                self.buffer = []

                # client loop
                try:
                    while not ws.closed:
                        msg = await ws.receive()
                        if msg.type == WSMsgType.CLOSE:
                            break
                        elif msg.type == WSMsgType.ERROR:
                            raise SocketError(ws.exception())
                        elif msg.type == WSMsgType.BINARY:
                            raise SocketError('Unexpected binary message')
                        elif msg.type == WSMsgType.TEXT:
                            event = msg.json()
                            if self.rpc.intercept_event(**event):
                                continue
                            await self.emit(**event, conn=self)

                    await self.emit(ON_CLOSE, conn=self)

                except SocketError as e:
                    await self.emit(ON_ERROR, conn=self, error=str(e))

                finally:
                    self.ws = None

    async def close(self):
        if self.connected:
            await self.ws.close()
        self.ws = None

    async def send(self, msg: dict) -> None:
        msg['ts'] = datetime.now().isoformat()
        if 'type' not in msg:
            raise Exception('Messages must have a type field')

        if not self.connected:
            # buffer messages while disconnected
            self.buffer.append(msg)
            return

        await self.ws.send_json(msg)



from aiohttp import web


class TaskAPI(object):
    def __init__(self, agent):
        self.agent = agent
        self.tasks = agent.tasks

    def routes(self, path: str = 'tasks'):
        return [
            web.post(path, self.create_task),
            web.get(path, self.get_tasks),
            web.get('%s/{id}' % path, self.get_task),
        ]

    async def get_tasks(self, req):
        tasks = list(self.tasks.values())
        return web.json_response(tasks)

    async def get_task(self, req):
        id = req.match_info['id']
        if id not in self.tasks:
            return web.HTTPNotFound()

        return web.json_response(self.tasks[id])

    async def create_task(self, req):
        taskdef = await req.json()
        task = self.agent.spawn(**taskdef)
        return web.json_response(task.serialize())



from cowait.tasks.shell import ShellTask


class SparkWorker(ShellTask):
    async def run(self, master: str, cores: int = 2, memory = '4G') -> dict:
        return await super().run(
            command=' '.join([
                'spark-class',
                'org.apache.spark.deploy.worker.Worker',
                master,
            ]),
            env={
                'SPARK_WORKER_CORES': str(cores),
                'SPARK_WORKER_MEMORY': memory,
            },
        )



from cowait.tasks import TaskDefinition
from cowait.engine.const import ENV_TASK_DEFINITION, MAX_ENV_LENGTH
from cowait.engine.utils import env_unpack, base_environment
from cowait.engine.errors import ProviderError


def create_ports(taskdef):
    return taskdef.ports


def create_env(cluster, taskdef):
    env = base_environment(cluster, taskdef)

    # check total length of environment data
    length = 0
    for key, value in env.items():
        length += len(str(key)) + len(str(value))

    if length > MAX_ENV_LENGTH:
        raise ProviderError(f'Task environment too long. Was {length}, max: {MAX_ENV_LENGTH}')

    return env


def extract_container_taskdef(container) -> TaskDefinition:
    for env in container.attrs['Config']['Env']:
        if ENV_TASK_DEFINITION == env[0:len(ENV_TASK_DEFINITION)]:
            data = env[len(ENV_TASK_DEFINITION)+1:]
            return TaskDefinition(**env_unpack(data))
    raise Exception('Unable to unpack container task definition')




# flake8: noqa: F401

from asyncio import sleep

from .status import WORK, WAIT, DONE, STOP, FAIL
from .messages import TASK_INIT, TASK_STATUS, TASK_LOG, TASK_FAIL, TASK_RETURN
from .errors import TaskError, TaskNotFoundError, StoppedError

from .task import Task
from .remote_task import RemoteTask
from .definition import TaskDefinition
from .instance import TaskInstance
from .decorator import task, spawn, input, exit
from .ops import join, wait

from .components.rpc import rpc



import inspect
import traceback
from aiohttp import web
from cowait.types import typed_async_call

RPC_CALL = 'rpc/call'
RPC_ERROR = 'rpc/error'
RPC_RESULT = 'rpc/result'


class RpcComponent():
    def __init__(self, task):
        self.task = task
        self.methods = get_rpc_methods(task)

        # listen for websocket rpc
        task.node.parent.on(RPC_CALL, self.on_rpc)
        task.node.server.on(RPC_CALL, self.on_rpc)

        # register http handler
        task.node.server.add_post('/rpc/{method}', self.http_rpc_handler)

    def get_method(self, method: str) -> callable:
        if method not in self.methods:
            raise RpcError(f'No such method {method}')

        return self.methods[method]

    async def call(self, method, args):
        rpc_func = self.get_method(method)
        result, _ = await typed_async_call(rpc_func, args)
        return result

    async def on_rpc(self, conn, method, args, nonce):
        try:
            rpc_func = self.get_method(method)

            result, result_type = await typed_async_call(rpc_func, args)

            await conn.send({
                'type': RPC_RESULT,
                'id': self.task.id,
                'nonce': nonce,
                'method': method,
                'args': args,
                'result': result,
                'result_type': result_type.describe(),
            })

        except Exception as e:
            traceback.print_exc()
            await conn.send({
                'type': RPC_ERROR,
                'id': self.task.id,
                'nonce': nonce,
                'method': method,
                'args': args,
                'error': str(e),
            })

    async def http_rpc_handler(self, req):
        try:
            call = await req.json()
            args = call.get('args', {})
            method = req.match_info['method']
            result = await self.call(method, args)
            return web.json_response({
                'success': True,
                'result': result,
            })
        except Exception as e:
            print('HTTP RPC Error:')
            traceback.print_exc()
            return web.json_response(
                {
                    'success': False,
                    'error': str(e),
                },
                status=400
            )


class RpcError(RuntimeError):
    pass


def rpc(f):
    """ Decorator for marking RPC methods """
    setattr(f, 'rpc', True)
    return f


def is_rpc_method(object):
    """ Returns true if the given object is a method marked with @rpc """
    if not inspect.ismethod(object):
        return False
    return hasattr(object, 'rpc')


def get_rpc_methods(object):
    """ Returns a dict of functions marked with @rpc on the given object """
    methods = inspect.getmembers(object, is_rpc_method)
    return {name: func for name, func in methods}



from datetime import datetime, timezone
from cowait.tasks import Task, TaskDefinition, sleep, rpc
from cowait.network import Conn, get_local_url
from cowait.tasks.status import WAIT, WORK, STOP
from cowait.tasks.messages import TASK_INIT, TASK_STATUS
from .tasklist import TaskList
from .subscriptions import Subscriptions
from .api import Dashboard, TaskAPI


class Agent(Task):
    def init(self):
        self.tasks = TaskList(self)
        self.subs = Subscriptions(self)

        # subscriber events
        async def send_state(conn: Conn) -> None:
            """" sends the state of all known tasks """
            for task in self.tasks.values():
                await conn.send({
                    'type': TASK_INIT,
                    'id':   task.id,
                    'task': task.serialize(),
                })
        self.subs.on('subscribe', send_state)

        self.token = self.meta['http_token']
        if self.token is None or self.token == '':
            self.node.server.auth.enabled = False
            self.token = 'none'

        # create http server
        self.node.server.add_routes(TaskAPI(self).routes('/api/1/tasks'))
        self.node.server.add_routes(Dashboard().routes())
        self.node.server.auth.add_token(self.token)

    async def run(self, **inputs) -> dict:
        if '/' in self.taskdef.routes:
            url = self.taskdef.routes['/']['url']
            print('Agent ready. Dashboard available at:')
            print(f'{url}?token={self.token}')
        else:
            print('Warning: No route set for /, dashboard not available.')

        # monitor tasks
        while True:
            running_tasks = self.cluster.list_all()
            for id, task in self.tasks.items():
                # consider only tasks that are waiting or running
                if task.status != WAIT and task.status != WORK:
                    continue

                # check if parent is dead
                if task.parent is not None and task.parent not in running_tasks:
                    parent = self.tasks[task.parent]
                    virtual_parent = parent.meta.get('virtual', False)
                    if not virtual_parent or parent.status != WORK:
                        since = datetime.now(timezone.utc) - task.created_at
                        await self.subtasks.emit_child_error(id, f'Task lost parent after {since}')

                        # destroy it
                        self.cluster.destroy(task.id)

                # special case for virtual tasks: since they dont exist as containers,
                # we cant check if they are still alive.
                if task.meta.get('virtual', False):
                    continue

                # ensure task is still in the list of running tasks
                # if not, consider it lost.
                if id not in running_tasks:
                    # compute task age
                    since = datetime.now(timezone.utc) - task.created_at
                    error = f'Task lost after {since}'
                    await self.subtasks.emit_child_error(id, error)

            await sleep(5.0)

        return {}

    @rpc
    async def destroy(self, task_id) -> None:
        await self.emulate_stop(task_id)
        self.cluster.destroy(task_id)

    @rpc
    async def destroy_all(self) -> None:
        for id, task in self.tasks.items():
            if task.status == WAIT or task.status == WORK:
                await self.emulate_stop(id)
        self.cluster.destroy_all()

    @rpc
    async def list_tasks(self) -> list:
        return self.cluster.list_all()

    @rpc
    async def get_agent_url(self) -> str:
        url = get_local_url()
        return f'{url}?token={self.token}'

    @rpc
    async def spawn(
        self,
        name: str,
        image: str,
        id: str = None,
        ports: dict = {},
        routes: dict = {},
        inputs: dict = {},
        meta: dict = {},
        env: dict = {},
        cpu: str = '0',
        memory: str = '0',
        owner: str = '',
        **kwargs: dict,
    ) -> dict:
        if not isinstance(name, str) and issubclass(name, Task):
            name = name.__module__

        # todo: throw error if any input is a coroutine

        task = self.cluster.spawn(TaskDefinition(
            id=id,
            name=name,
            image=image,
            upstream=get_local_url(),
            meta=meta,
            ports=ports,
            routes=routes,
            env=env,
            cpu=cpu,
            memory=memory,
            owner=owner,
            inputs={
                **inputs,
                **kwargs,
            },
        ))

        # authorize id
        self.node.server.auth.add_token(id)

        # register with subtask manager
        self.subtasks.watch(task)

        return task.serialize()

    async def emulate_stop(self, task_id: str):
        await self.node.server.emit(type=TASK_STATUS, id=task_id, status=STOP, conn=None)



from __future__ import annotations
from .status import WAIT
from .definition import TaskDefinition


class TaskInstance(TaskDefinition):
    def __init__(
        self,
        status: str = WAIT,
        error: str = None,
        result: any = None,
        log: str = '',
        **data,
    ):
        super().__init__(**data)
        self.status = status
        self.error = error
        self.result = result
        self.log = log

    def serialize(self) -> dict:
        """ Serialize task definition to a dict """
        taskdef = super().serialize()
        return {
            **taskdef,
            'status': self.status,
            'error': self.error,
            'result': self.result,
            'log': self.log,
        }

    @staticmethod
    def deserialize(instance: dict) -> TaskInstance:
        """ Deserialize task definition from a dict """
        taskdef = TaskDefinition.deserialize(instance).serialize()
        return TaskInstance(**{
            **taskdef,
            'status': instance.get('status', WAIT),
            'error': instance.get('error', None),
            'result': instance.get('result', None),
            'log': instance.get('log', None),
        })

